from pathlib import Path
from typing import Union, List, Optional

import requests
import toml
from apscheduler.schedulers.blocking import BlockingScheduler
from pydantic import ValidationError

from think_dashboard_agent import types
from think_dashboard_agent.providers import DEFAULT_PROVIDERS


def __check_instances(instances: List, providers: Optional[dict] = None):
    _data = []

    if providers is None:
        providers = DEFAULT_PROVIDERS
    else:
        providers = {**DEFAULT_PROVIDERS, **providers}

    for instance in instances:
        provider = providers[instance.type]
        _instance_check_result = provider(instance).check()
        _data.append({
            'name': instance.name,
            'key': instance.key,
            'type': instance.type,
            'response': _instance_check_result.dict(),
        })
    return _data


def __run(session: requests.Session, config: types.Config, providers: Optional[dict] = None):
    _data = __check_instances(config.instance, providers)
    print(_data)
    # r = session.post(config.api_key, json=_data)
    # if r.status_code != 200:
    #     print(r.json())


def run(config_file: Union[str, Path], providers: Optional[dict] = None):
    """
    Run the agent.
    """
    with open(config_file) as f:
        _toml_file = toml.load(f)
    try:
        config = types.Config(**_toml_file)
        with requests.Session() as session:
            session.headers.update({'Authorization': f"Token {config.api_key}"})
            __run(session, config, providers)

    except ValidationError as e:
        print(e)


def run_forever(config_file: Union[str, Path], interval: int, providers: Optional[dict] = None):
    """
    Run the agent forever.
    """
    scheduler = BlockingScheduler()
    with open(config_file) as f:
        _toml_file = toml.load(f)
    try:
        config = types.Config(**_toml_file)
        with requests.Session() as session:
            session.headers.update({'Authorization': f"Token {config.api_key}"})
            scheduler.add_job(__run, 'interval', [session, config, providers], seconds=interval)
            scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        scheduler.shutdown()



import psycopg2

from think_dashboard_agent.types import DatabaseInstance
from .base import BaseProvider
from ..responses import InstanceCheckResult


class DatabaseProvider(BaseProvider):
    def __init__(self, instance: DatabaseInstance):
        self.instance = instance
        self.conn = None

    def exc(self):
        cur = self.conn.cursor()
        cur.execute("""
                    SELECT count(table_name) 
                        FROM information_schema.tables
                        WHERE table_schema = 'public'
                    """)
        return cur.fetchone()[0]

    def check(self):
        try:
            self.connect()
            result = self.exc()
            return InstanceCheckResult(data=result)
        except psycopg2.OperationalError:
            return InstanceCheckResult(status=500, error='Unable to connect to the database')
        except psycopg2.Error as e:
            return InstanceCheckResult(status=500, error=str(e))
        finally:
            if self.conn:
                try:
                    self.conn.close()
                except psycopg2.Error:
                    pass

    # Create a connection to the database
    def connect(self):
        conn = psycopg2.connect(
            host=self.instance.host,
            port=self.instance.port,
            user=self.instance.username,
            password=self.instance.password,
            database=self.instance.database
        )
        self.conn = conn

    # Close the connection to the database
    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None



import redis

from think_dashboard_agent.responses import InstanceCheckResult
from think_dashboard_agent.types import RedisInstance
from .base import BaseProvider


class RedisProvider(BaseProvider):
    def __init__(self, instance: RedisInstance):
        self.instance = instance
        self.conn = None

    def connect(self):
        conn = redis.Redis(
            host=self.instance.host,
            port=self.instance.port,
            password=self.instance.password,
            db=self.instance.db
        )
        self.conn = conn

    def exc(self):
        return self.conn.execute_command('PING')

    def check(self) -> InstanceCheckResult:
        try:
            self.connect()
            return InstanceCheckResult(data=self.exc())
        except redis.RedisError as e:
            return InstanceCheckResult(status=500, data={'error': str(e)})
        finally:
            self.close()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None


import requests

from think_dashboard_agent.types import ElasticSearchInstance
from .base import BaseProvider
from ..responses import InstanceCheckResult


class ElasticSearchProvider(BaseProvider):
    def __init__(self, instance: ElasticSearchInstance):
        self.instance = instance
        self.conn = None

    def connect(self):
        self.conn = requests.Session()
        self.conn.auth = (self.instance.username, self.instance.password)
        self.conn.headers.update({'Content-Type': 'application/json'})

    def exc(self):
        return self.conn.get(f"{self.instance.host}:{self.instance.port}/_cluster/health")

    def check(self) -> InstanceCheckResult:
        try:
            self.connect()
            r = self.exc()
            if r.status_code == 200:
                return InstanceCheckResult(status=200, data=r.json())
            else:
                return InstanceCheckResult(status=r.status_code, error=r.text)
        except requests.exceptions.RequestException as e:
            return InstanceCheckResult(status=500, error=str(e))
        finally:
            self.close()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None



from typing import Union

from pydantic import BaseModel


class InstanceCheckResult(BaseModel):
    status: int = 200
    data: Union[Union[Union[dict, list], str], None] = None
    error: str = None



from think_dashboard_agent.responses import InstanceCheckResult


class BaseProvider:

    def connect(self):
        raise NotImplementedError()

    def exc(self, *args, **kwargs):
        raise NotImplementedError()

    def check(self) -> InstanceCheckResult:
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()



import socket
import ssl
from datetime import datetime

from think_dashboard_agent.providers.base import BaseProvider
from think_dashboard_agent.responses import InstanceCheckResult
from think_dashboard_agent.types import SSLInstance


class SSLProvider(BaseProvider):
    def __init__(self, instance: SSLInstance):
        self.instance = instance
        self.conn = None

    def connect(self):
        if self.conn is None:
            context = ssl.create_default_context()
            self.conn = context.wrap_socket(
                socket.socket(socket.AF_INET),
                server_hostname=self.instance.host,
            )
            self.conn.settimeout(3.0)
            self.conn.connect((self.instance.host, self.instance.port))

    def exc(self):
        ssl_date_fmt = r'%b %d %H:%M:%S %Y %Z'
        ssl_info = self.conn.getpeercert()
        expires = datetime.strptime(ssl_info['notAfter'], ssl_date_fmt)
        return expires.strftime('%Y-%m-%d %H:%M:%S')

    def check(self) -> InstanceCheckResult:

        try:
            self.connect()
            return InstanceCheckResult(status=200, data=self.exc())
        except Exception as e:
            return InstanceCheckResult(status=500, error=str(e))
        finally:
            self.close()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['think_dashboard_agent', 'think_dashboard_agent.providers']

package_data = \
{'': ['*']}

install_requires = \
['APScheduler>=3.9.1,<4.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'redis>=4.3.4,<5.0.0',
 'requests>=2.28.1,<3.0.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'think-dashboard-agent',
    'version': '1.0.0',
    'description': '',
    'long_description': None,
    'author': 'Suhrob Malikov',
    'author_email': 'suhrob@thinkland.uz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import List, Literal, Dict

from pydantic import BaseModel, validator

from think_dashboard_agent.exceptions import ConfigParseException


class BaseInstance(BaseModel):
    key: str
    name: str = None
    type: Literal['database', 'redis', 'elasticsearch', 'ssl']
    host: str
    port: int


class DatabaseInstance(BaseInstance):
    username: str
    password: str
    database: str
    port: int = 5432


class RedisInstance(BaseInstance):
    password: str = None
    db: int = 1


class ElasticSearchInstance(BaseInstance):
    username: str
    password: str

    @validator('host')
    def validate_host(cls, v):
        if not v.startswith('http'):
            v = 'https://{}'.format(v)
        return v


class SSLInstance(BaseInstance):
    port: int = 443

    @validator('host')
    def validate_host(cls, v):
        if v.startswith('http://'):
            v = v.replace('http://', '')
        elif v.startswith('https://'):
            v = v.replace('https://', '')
        return v


INSTANCE_TYPES = {
    'database': DatabaseInstance,
    'redis': RedisInstance,
    'elasticsearch': ElasticSearchInstance,
    'ssl': SSLInstance,
}
ALLOWED_TYPES = list(INSTANCE_TYPES.keys())


class Config(BaseModel):
    api_key: str
    dashboard_url: str
    server_name: str
    instance: List[Dict]

    @validator('instance')
    def validate_instances(cls, v):
        _instances = []
        _keys = set()
        for instance in v:
            if 'type' not in instance:
                raise ConfigParseException('Instance must have a type')

            _instance_type = INSTANCE_TYPES[instance['type']]
            _instance = _instance_type(**instance)
            _instances.append(_instance)
            _keys.add(_instance.key)

        if len(_keys) != len(_instances):
            raise ConfigParseException('All instances must have a unique key')
        return _instances



from .database import DatabaseProvider
from .elasticsearch import ElasticSearchProvider
from .redis_provider import RedisProvider
from .ssl_provider import SSLProvider

DEFAULT_PROVIDERS = {
    'database': DatabaseProvider,
    'redis': RedisProvider,
    'elasticsearch': ElasticSearchProvider,
    'ssl': SSLProvider
}

__all__ = [
    'DatabaseProvider',
    'RedisProvider',
    'ElasticSearchProvider',
    'SSLProvider',
    'DEFAULT_PROVIDERS'
]
